"""
File: metrics.py 
-------------- 
Jax-based metrics. 
""" 

import jax.numpy as jnp


class MeanMetric:
  """
  Base scalar metric designed to use on a per-epoch basis 
  and updated on per-batch basis. For getting average across
  the epoch. 
  """

  def __init__(self):
    self.vals = []

  def update(self, new_val):
    # if isinstance(new_val, jnp.array): # squeeze only when true? 
    #     pass 

    new_val = jnp.squeeze(new_val)  # remove any excess dimensions 
    self.vals.append(new_val)

  def reset(self):
    self.vals = []

  def get(self):
    mean_value = sum(self.vals)/len(self.vals)
    return 


class Accuracy(MeanMetric):
    """
    Subclass of MeanMetric which defines 
    a standard accuracy update function. 
    """

    def update(self, logits, labels):
        new_val = jnp.squeeze(jnp.mean(jnp.argmax(logits, -1) == labels))
        self.vals.append(new_val)



def compute_accuracy(logits, labels):
  """stand alone accuracy metric computation"""
  return jnp.mean(jnp.argmax(logits, -1) == labels)



def compute_binary_accuracy(logits, targets):
  """
  Computes stand alone accuracy for binary classification
  where the net outputs one element per prediction. 
  Source: https://github.com/tristandeleu/jax-meta-learning/blob/master/jax_meta/utils/metrics.py
  """

  return jnp.mean((logits > 0).astype(targets.dtype) == targets)
  


"""
File: ml.py 
-------------- 
Contains useful functions and utilities for machine learning in Jax. 
""" 


import jax
from jax import grad, jit, vmap, tree_util
import jax.numpy as jnp


# --------------------------- (Losses) ------------------------------

bce_loss = lambda logits, y: -(y * jax.nn.log_sigmoid(logits) + (1 - y) * jax.nn.log_sigmoid(1-logits)) 
bce_batched = vmap(bce_loss)


# --------------------------- (Training) ------------------------------
# gradient_step = lambda p, g, lr: p - lr * g  # doesn't work because lr needs to be a pytree then 

gradient_step = lambda p, g: p - lr * g  # lr needs to be defined in user scope 


@jax.jit
def update_step(params, grads, lr):
    """
    Updates params via one step of gradient descent. 
    Arguments: 
    -----------
    - params: pytree of parameters for the net 
    - grads: gradient of loss wrt to params 
    - lr: learning rate (step size) 
    Returns: 
    -----------
    loss_value: updated params  
    """
    gradient_step = lambda p, g: p - lr * g   
    return tree_util.tree_map(gradient_step, params, grads)


def stateless_loss(params, net_apply, loss_function, x, y, return_logits=False):
    """
    Computes loss in single function using arguments in stateless (no OOP). 
    Useful to wrap this function in 'grad' to get the gradient with respect
    to params. 
    Arguments: 
    -----------
    - params: pytree of parameters for the net 
    - net_apply: model apply fn 
    - loss_function: function for computing the loss 
    - x: input
    - y: label 
    - return_logits: if True, returns logits as well (must specify hax_aux=True when taking grad then)
    Returns: 
    -----------
    loss_value: scalar loss value 
    """

    model_preds = net_apply(params, x)
    loss_value = loss_function(model_preds, y)
    loss_value = jnp.squeeze(loss_value)  # remove axes 
    if return_logits == False:
        return loss_value
    else:
        loss_value, model_preds



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['easyjax']

package_data = \
{'': ['*']}

install_requires = \
['jax==0.3.15']

setup_kwargs = {
    'name': 'easyjax',
    'version': '0.1.1',
    'description': 'Facilitates machine learning development for JAX.',
    'long_description': '# EasyJax \n\n> work-in-progress. \n\nEasyJax is a python package that facilitates machine learning development for JAX. It does that by providing:\n\n1. A high-level API for machine learning workflows in JAX (specifically a trainer, experiment parent class). \n2. Several machine learning specific utilities for working with JAX (e.g., `ml.update_step`). \n\n',
    'author': 'Rohan Sikand',
    'author_email': 'rsikand29@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/rosikand/easyjax',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
